#include "emu.h"

/* --------------------------------------------------------------------- */

const char *feature_bits[32] = {
    [ KVM_FEATURE_CR3_CACHE ] = "cr3-cache",
};

void pv_init(struct xen_cpu *cpu)
{
    char buf[128];
    struct kvm_cpuid_entry entry;
    uint32_t sig[3];
    uint32_t features;

    entry.function = KVM_CPUID_SIGNATURE;
    real_cpuid(&entry);
    sig[0] = entry.ebx;
    sig[1] = entry.ecx;
    sig[2] = entry.edx;

    entry.function = KVM_CPUID_FEATURES;
    real_cpuid(&entry);
    features = entry.eax;

    snprintf(buf, sizeof(buf), "%s: cpu %d, signature \"%.12s\", features 0x%08x",
	     __FUNCTION__, cpu->id, (char*)sig, features);
    print_bits(1, buf, features, features, feature_bits);

#if 0
    /* This is work-in-progress, don't enable this unless
     * you know what you are doing. */
    if (features & (1 << KVM_FEATURE_CLOCKSOURCE)) {
	uint64_t clock = EMU_PA(&cpu->vcpu_info->time);
	printk("%s: register vcpu %d pv clock at 0x%" PRIx64 "\n",
	       __FUNCTION__, vcpu, clock);
	if (0 != wrmsrl_safe(MSR_KVM_PARAVIRT_CLOCK, clock))
	    panic("pv clock wrmsr failed");
    }
#endif

#if 0
    /* This is work-in-progress, don't enable this unless
     * you know what you are doing. */
    if (1 /* features & (1 << KVM_FEATURE_CR3_CACHE) */) {
	uint64_t cache;
	cpu->cr3_cache = get_pages(1, "cr3 cache");
//	cache = EMU_PA(cpu->cr3_cache);    /* physical */
	cache = (uintptr_t)cpu->cr3_cache; /* virtual  */
	printk(1, "%s: register cr3 cache at 0x%" PRIx64 " ...\n",
	       __FUNCTION__, cache);
	if (0 != wrmsrl_safe(KVM_MSR_SET_CR3_CACHE, cache)) {
	    printk(1, "%s: ... FAILED\n", __FUNCTION__);
	} else {
	    printk(1, "%s: ... OK, %d entries\n",
		   __FUNCTION__, cpu->cr3_cache->max_idx);
	}
    }
#endif
}

/* --------------------------------------------------------------------- */

void pv_write_cr3(struct xen_cpu *cpu, ureg_t cr3)
{
    int idx;

    if (cpu->cr3_cache && cpu->cr3_cache->max_idx) {
	for (idx = 0; idx < cpu->cr3_cache->max_idx; idx++) {
	    if (cpu->cr3_cache->entry[idx].guest_cr3 == cr3) {
		/*
		 * Cache-hit: we load the cached host-CR3 value.
		 * This never causes any VM exit. (if it does then the
		 * hypervisor could do nothing with this instruction
		 * and the guest OS would be aborted)
		 */
		vminfo.faults[XEN_FAULT_OTHER_CR3_CACHE_HIT]++;
		write_cr3(cpu->cr3_cache->entry[idx].host_cr3);
		return;
	    }
	}
    }
    write_cr3(cr3);
    return;
}
